/*
 * Implementaco de um filtro em grade multi-cannal em que os coeficiente do filtro 
 * são os mesmos que os de outra lattice.
 *
 * As entradas sao:
 * - u(n), reference signals
 * - d(n), desired signals
 *
 * As saidas sao:
 * d^(n|Un)
 * o que é calculado basedo no erro de predição apriori:
 * d^(n|Un) = d(n) - gamma(n-1)*alfa1(n)
 *
 * Paulo Lopes 17/12/2001
 */

#include "global.h"
#include <math.h>
#include <memory.h> 
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "McLatticePr.h"

McLatticePr::McLatticePr(int inSizeAr, int outSizeAr, int orderAr,  
				 MyData *GammaFAr, MyData *GammaBAr, MyData *kappa0Ar, MyData *gamma0Ar)
{
	IniLatticePr(inSizeAr, outSizeAr, orderAr);

	GammaF0p=GammaFAr;
	GammaB0p=GammaBAr;
	kappa0p=kappa0Ar;
	gamma0p=gamma0Ar;

	link= true;
}

void McLatticePr::IniLatticePr(int inSizeAr, int outSizeAr, int orderAr)
{
	int m,i;

	order=orderAr;
	inSize=inSizeAr;
	outSize=outSizeAr;
	inSizeSq=inSize*inSize;
	outSizeSq=outSize*outSize;

	eta1p=(MyData *)calloc(inSize*(order+1), sizeof(MyData));
	psi1p=(MyData *)calloc(inSize*(order+1), sizeof(MyData));
	alfa1p=(MyData *)calloc(outSize*(order+1), sizeof(MyData));

	eta0p=(MyData *)calloc(inSize*(order+1), sizeof(MyData));
	psi0p=(MyData *)calloc(inSize*(order+1), sizeof(MyData));
	alfa0p=(MyData *)calloc(outSize*(order+1), sizeof(MyData));

	for(m=0; m<=order; m++) {
		for(i=0; i<inSize; i++)
			eta0(m,i)=0;
		for(i=0; i<inSize; i++)
			psi0(m,i)=0;
		for(i=0; i<outSize; i++)
			alfa0(m,i)=0;
	}
}

McLatticePr::McLatticePr(int inSize, int outSize, int order)
{
	int m,i,j;

	IniLatticePr(inSize, outSize, order);

	GammaF0p=(MyData *)calloc(inSizeSq*(order+1), sizeof(MyData));
	GammaB0p=(MyData *)calloc(inSizeSq*(order+1), sizeof(MyData));
	kappa0p=(MyData *)calloc(outSize*inSize*(order+1), sizeof(MyData));
	gamma0p=(MyData *)calloc(order+1, sizeof(MyData));

	for(m=0; m<=order; m++) {

		for(i=0; i<inSize; i++){
			for(j=0; j<inSize; j++)
				GammaF0(m,i,j)=0;}
		for(i=0; i<inSize; i++){
			for(j=0; j<inSize; j++)
				GammaB0(m,i,j)=0;}
		gamma0(m)=1;
		for(i=0; i<outSize; i++) {
			for(j=0; j<inSize; j++)
				kappa0(m,i,j)=0;
		}
	}

	link = false;
}

McLatticePr::~McLatticePr(void)
{
	free(eta1p);
	free(psi1p);
	free(alfa1p);

	free(eta0p);
	free(psi0p);
	free(alfa0p);

	if (!link) {
		free(GammaF0p);
		free(GammaB0p);
		free(gamma0p);
		free(kappa0p);
	}
}



// The implementation of the lattice copy has simetric coefficients
void McLatticePr::Update(val_type *in)
{
	int m,i,k;

	// Inicialização
	
	for(i=0; i<inSize; i++)
		eta1(0,i)=(MyData)in[i];

	for(i=0; i<inSize; i++)
		psi1(0,i)=(MyData)in[i];

	for(i=0; i<outSize; i++)
		alfa1(0,i)=(MyData)(0.0);

	for(m=1; m<=order; m++){

		// eta1[m] = eta1[m - 1] + GammaF0[m].psi0[m - 1];

		for(i=0; i<inSize; i++) {
			eta1(m,i)=eta1(m-1,i);
			for(k=0; k<inSize; k++)
				eta1(m,i)+=GammaF0(m,i,k)*psi0(m-1,k);
		}
	
		// psi1[m] = psi0[m - 1] + GammaB0[m].eta1[m - 1];

		for(i=0; i<inSize; i++) {
			psi1(m,i)=psi0(m-1,i);
			for(k=0; k<inSize; k++)
				psi1(m,i)+=GammaB0(m,i,k)*eta1(m-1,k);
		}
		
		
		// alfa1[m]  = alfa1[m - 1] - kappa0[m].psi1[m - 1]
		// the minus is replaced by a sum

		for(i=0; i<outSize; i++) {
			alfa1(m,i)=alfa1(m-1,i);
			for(k=0; k<inSize; k++)
				alfa1(m,i)+=kappa0(m,i,k)*psi1(m-1,k);
		}
		
	}

	// time update

	for(m=0; m<=order; m++){
		
		for(i=0; i<inSize; i++)
			eta0(m,i)=eta1(m,i);
		
		for(i=0; i<inSize; i++)
			psi0(m,i)=psi1(m,i);
		
		for(i=0; i<outSize; i++)
			alfa0(m,i)=alfa1(m,i);
	}
}

void McLatticePr::Output(val_type *out)
{
	int i;
	for(i=0; i<outSize; i++)
		// out[i]=gamma0(order)*alfa1(order,i); this is the a posteriori error
		out[i]=alfa1(order,i);
}
